/*
** Module   : spawn.c
** Abstract : a simple tool able to spawn a process on an existing user account.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 MAG 20132711 First version.
*/

#define _GNU_SOURCE
#include <crypt.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <shadow.h>
#include <signal.h>
#include <grp.h>
//#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1

/* Global data */
struct passwd *pw;

/**
* Assign a controlling terminal to a process.
*
* @return  0 - Success
*          1 - Error
*/
int assign_pty()
{
    int fdm, fds, rc;

    /* Open a pseudo-terminal device */
    fdm = posix_openpt(O_RDWR);
    if (fdm < 0)
    {
#ifdef DEBUG
        perror("Error on posix_openpt().\n");
#endif
        return EXIT_FAILURE;
    }

    /* Grant access to the slave pseudo-terminal device */
    rc = grantpt(fdm);
    if (rc != 0)
    {
#ifdef DEBUG
        perror("Error on grantpt().\n");
#endif
        return EXIT_FAILURE;
    }

    /*  Unlock a pseudo-terminal master/slave pair */
    rc = unlockpt(fdm);
    if (rc != 0)
    {
#ifdef DEBUG
        perror("Error on unlockpt().\n");
#endif
        return EXIT_FAILURE;
    }

    /* open slave pseudo-terminal device */
    fds = open(ptsname(fdm), O_RDWR);
    if (fds < 0)
    {
#ifdef DEBUG
        perror("Error on open slave device.\n");
#endif
        return EXIT_FAILURE;
    }

    /* Use the slave as TTY */
    if (ioctl(fds, TIOCSCTTY, (char *)1) == -1)
    {
#ifdef DEBUG
        perror("Error on ioctl(TIOCSCTTY)\n");
#endif
        return EXIT_FAILURE;
    }

    /* Set stdin, stdout and stderr */
    dup2(STDIN_FILENO, fds); // PTY becomes standard input (0)
    dup2(STDOUT_FILENO, fds); // PTY becomes standard output (1)
    dup2(STDERR_FILENO, fds); // PTY becomes standard error (2)

    /* Cancel certain signals */
    signal(SIGCHLD,SIG_DFL); /* A child process dies */
    signal(SIGTSTP,SIG_IGN); /* Various TTY signals */
    signal(SIGTTOU,SIG_IGN);
    signal(SIGTTIN,SIG_IGN);
    signal(SIGHUP, SIG_IGN); /* Ignore hangup signal */
    signal(SIGTERM,SIG_DFL); /* Die on SIGTERM */

    return EXIT_SUCCESS;
}

/**
* Execute a command on a specified user account.
* A child process is created via fork().
* The child process uid and gid are set as the new user uid and gid.
* Create a new session for child. The child became a session leader.
* Assign a controlling terminal to session leader.
* The child process environment is set from the new user account.
* Create a child process tu execute the command.
*
* @param   command
*          Command to execute
* @param   args
*          Command arguments.

* @return  0
*/
int spawn(const char* command, char* const argv[])
{
    pid_t child_pid, proc_pid;
    int status;

    /* Creates a child process */
    child_pid = fork();

    if (child_pid == 0)
    {
        /* Set gid and egit*/
        if (setgid(pw->pw_gid) < 0  || setegid(pw->pw_gid) < 0 || getgid() != pw->pw_gid || getegid() != pw->pw_gid)
        {
            exit(EXIT_FAILURE);
        }

        /* Initialize the supplementary group access list */
        if (initgroups(pw->pw_name, pw->pw_gid) < 0)
        {
            exit(EXIT_FAILURE);
        }

        /* Set uid and euid */
        if (setuid(pw->pw_uid) < 0 || seteuid(pw->pw_uid) < 0 || getuid() != pw->pw_uid || geteuid() != pw->pw_uid)
        {
            exit(EXIT_FAILURE);
        }

        /* Creates a session and sets the process group ID */
        if(setsid() < 0)
        {
            exit(EXIT_FAILURE);
        }

        /* Assign controlling terminal */
        if (assign_pty() != 0)
        {
            exit(EXIT_FAILURE);
        }

        /* Set user environment */
        char* path = strdup(getenv("PATH"));
        setenv("HOME",pw->pw_dir,1);
        setenv("USER",pw->pw_name,1);
        setenv("LOGNAME",pw->pw_name,1);
        setenv("LOGIN",pw->pw_name,1); /*< Historical; only strictly needed on AIX */
        setenv("SHELL",pw->pw_shell[0] ? pw->pw_shell : "/bin/sh",1);
        setenv("PATH",path,1);
        free(path);

        /* Change the current working directory */
        if (chdir(getenv("HOME")) < 0)
        {
            if (chdir("/") < 0)
            {
#ifdef DEBUG
                perror("chdir($HOME)\n");
#endif
                exit(EXIT_FAILURE);
            }
        }

        /* Spawn a child process */
        proc_pid = fork();

        if (proc_pid == 0)
        {
            if (setpgid(0, child_pid) < 0)
            {
#ifdef DEBUG
                perror("setpgid()\n");
#endif
                exit(EXIT_FAILURE);
            }

            /* Execute command */
            exit(execvp(command, argv));
        }
        else if (proc_pid < 0)
        {
            /* On fork() fails */
            exit(EXIT_FAILURE);
        }

        /* Wait for child to exit*/
        wait(&status);

        /* Exit */
        exit(status);
    }
    else if (child_pid < 0)
    {
        /* On fork() fails */
        exit(EXIT_FAILURE);
    }

    return 0;
}

/**
* Get user account informations including uid, gid and password.
* If the user exists check for password match.
* This is basically an authentication action.
* The user root is not allowed.
*
* @param   username
*          User name.
* @param   password
*          Use password.

* @return  0 - success.
*          1 - user does not exists
*          2 - password doe not match.
*          3 - user is root.
*/
int check_user (const char* username, const char* password)
{
    struct spwd *sp;
    char *encrypted, *correct;

    pw = getpwnam(username);
    endpwent();

    if (!pw) return 1; //user doesn't really exist

    if (pw->pw_uid == 0)
    {
        return 3;   // user is root
    }

    sp = getspnam(pw->pw_name);
    endspent();

    correct = sp ? sp->sp_pwdp : pw->pw_passwd;
    encrypted = crypt(password, correct);

    return strcmp(encrypted, correct) ? 2 : 0;  // bad pw=2, success=0
}

/**
* Check executable file attributes.
*
* @param   path
*          path to executable file.
*
* @return  0 - OK.
*          1 - No root permissions.
*/
int chek_root(const char *path)
{
    struct stat pstat;
    int result;

    result = stat(path, &pstat);

    if (!result)
    {
        result = pstat.st_uid == 0 &&
                 pstat.st_gid == 0 &&
                 (pstat.st_mode & ALLPERMS) == 04755 ? 0 : 1;
    }

    return result;
}

/**
* Get the user password from stdin.
*
* @return  Password or NULL on error.
*/
char* password()
{
    int read;
    size_t len;
    char* pswd = NULL;

    /* Read a line form console */
    read = getline(&pswd, &len, stdin);

    if (read == -1)
    {
        return NULL;
    }
    else if (pswd[read-1] == '\n')
    {
        pswd[read-1] = 0;
    }

    return pswd;
}

/**
* Main function.
* Link option: -lm -lcrypt
* After build do the following:
*     $ sudo -s
*     $ chown root:root spawn
*     $ chmod 4755 spawn
*
* @param   argc
*          Number of command line arguments.
* @param   argv
*          Command line arguments.
*
* @return  Exit status.
*/
int main(int argc, char *argv[])
{
    char *psw;
    int result;

    /* Check for root permissions */
    if (chek_root(argv[0]))
    {
        puts("No root access permissions.");
        exit(EXIT_FAILURE);
    }

    /* Print a short help */
    if (argc < 3)
    {
        puts("Usage: spawn <usr> <command> [args]");
        exit(EXIT_FAILURE);
    }

    /* Ask for password */
    psw = password();

    if(psw == NULL)
    {
        puts("No password.");
        exit(EXIT_FAILURE);
    }

    /* Check user account credentials */
    result = check_user(argv[1], psw);

    /* Password no longer needed. */
    free(psw);

    switch(result)
    {
        case 0:
            result = spawn(argv[2], &argv[2]);
            break;
        case 1:
            puts("Invalid user name.");
            break;
        case 2:
            puts("Invalid password.");
            break;
        case 3:
            puts("User root not allowed.");
            break;
    }

    exit(result);
}
